import os
import sys
from textwrap import dedent

import pytest

import salt.modules.pip as pip
import salt.utils.files
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

TARGET = []
if os.environ.get("VENV_PIP_TARGET"):
    TARGET = ["--target", os.environ.get("VENV_PIP_TARGET")]


class FakeFopen:
    def __init__(self, filename):
        d = {
            "requirements-0.txt": (
                b"--index-url http://fake.com/simple\n\n"
                b"one  # -r wrong.txt, other\n"
                b"two # --requirement wrong.exe;some\n"
                b"three\n"
                b"-r requirements-1.txt\n"
                b"# nothing\n"
            ),
            "requirements-1.txt": (
                "four\n"
                "five\n"
                "--requirement=requirements-2.txt\t# --requirements-2.txt\n\n"
            ),
            "requirements-2.txt": b"""six""",
            "requirements-3.txt": (
                b"# some comment\n"
                b"-e git+ssh://git.example.com/MyProject#egg=MyProject # the project\n"
                b"seven\n"
                b"-e git+ssh://git.example.com/Example#egg=example\n"
                b"eight # -e something#or other\n"
                b"--requirement requirements-4.txt\n\n"
            ),
            "requirements-4.txt": "",
        }
        self.val = d[filename]

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def read(self):
        return self.val


@pytest.fixture
def expected_user():
    return "fnord"


@pytest.fixture
def configure_loader_modules():
    return {pip: {"__salt__": {"cmd.which_bin": lambda _: "pip"}}}


def test__pip_bin_env():
    ret = pip._pip_bin_env(None, "C:/Users/ch44d/Documents/salt/tests/pip.exe")
    if salt.utils.platform.is_windows():
        assert ret == "C:/Users/ch44d/Documents/salt/tests"
    else:
        assert ret is None


def test__pip_bin_env_no_change():
    cwd = "C:/Users/ch44d/Desktop"
    ret = pip._pip_bin_env(cwd, "C:/Users/ch44d/Documents/salt/tests/pip.exe")
    assert ret == cwd


def test__pip_bin_env_no_bin_env():
    ret = pip._pip_bin_env(None, None)
    assert ret is None


@pytest.fixture
def python_binary():
    binary = [sys.executable, "-m", "pip"]
    if hasattr(sys, "RELENV"):
        binary = [str(sys.RELENV / "salt-pip")]
    return binary


def test_install_frozen_app(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch("sys.frozen", True, create=True):
        with patch("sys._MEIPASS", True, create=True):
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkg)
                expected = [
                    *python_binary,
                    "install",
                    *TARGET,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    python_shell=False,
                    saltenv="base",
                    use_vt=False,
                    runas=None,
                )


def test_install_source_app(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch("sys.frozen", False, create=True):
        with patch("sys._MEIPASS", False, create=True):
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkg)
                expected = [
                    *python_binary,
                    "install",
                    *TARGET,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    python_shell=False,
                    saltenv="base",
                    use_vt=False,
                    runas=None,
                )


def test_fix4361(python_binary):
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(requirements="requirements.txt")
        expected_cmd = [
            *python_binary,
            "install",
            "--requirement",
            "requirements.txt",
            *TARGET,
        ]
        mock.assert_called_with(
            expected_cmd,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_editable_without_egg_fails():
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(
            CommandExecutionError,
            pip.install,
            editable="git+https://github.com/saltstack/salt-testing.git",
        )


def test_install_multiple_editable(python_binary):
    editables = [
        "git+https://github.com/saltstack/istr.git@v1.0.1#egg=iStr",
        "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
    ]

    expected = [*python_binary, "install", *TARGET]
    for item in editables:
        expected.extend(["--editable", item])

    # Passing editables as a list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(editable=editables)
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing editables as a comma separated list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(editable=",".join(editables))
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_multiple_pkgs_and_editables(python_binary):
    pkgs = ["pep8", "salt"]
    editables = [
        "git+https://github.com/saltstack/istr.git@v1.0.1#egg=iStr",
        "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
    ]

    expected = [*python_binary, "install", *TARGET]
    expected.extend(pkgs)
    for item in editables:
        expected.extend(["--editable", item])

    # Passing editables as a list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkgs=pkgs, editable=editables)
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing editables as a comma separated list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkgs=",".join(pkgs), editable=",".join(editables))
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # As single string (just use the first element from pkgs and editables)
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkgs=pkgs[0], editable=editables[0])
        expected = [
            *python_binary,
            "install",
            *TARGET,
            pkgs[0],
            "--editable",
            editables[0],
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_issue5940_install_multiple_pip_mirrors(python_binary):
    """
    test multiple pip mirrors.  This test only works with pip < 7.0.0
    """
    with patch.object(pip, "version", MagicMock(return_value="1.4")):
        mirrors = [
            "http://g.pypi.python.org",
            "http://c.pypi.python.org",
            "http://pypi.crate.io",
        ]

        expected = [*python_binary, "install", "--use-mirrors"]
        for item in mirrors:
            expected.extend(["--mirrors", item])
        expected = [*expected, *TARGET, "pep8"]

        # Passing mirrors as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=["pep8"], mirrors=mirrors)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing mirrors as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=["pep8"], mirrors=",".join(mirrors))
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        expected = [
            *python_binary,
            "install",
            "--use-mirrors",
            "--mirrors",
            mirrors[0],
            *TARGET,
            "pep8",
        ]

        # As single string (just use the first element from mirrors)
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=["pep8"], mirrors=mirrors[0])
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_install_with_multiple_find_links(python_binary):
    find_links = [
        "http://g.pypi.python.org",
        "http://c.pypi.python.org",
        "http://pypi.crate.io",
    ]
    pkg = "pep8"

    expected = [*python_binary, "install"]
    for item in find_links:
        expected.extend(["--find-links", item])
    expected = [*expected, *TARGET, pkg]

    # Passing mirrors as a list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, find_links=find_links)
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing mirrors as a comma separated list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, find_links=",".join(find_links))
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Valid protos work?
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, find_links=find_links)
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    expected = [
        *python_binary,
        "install",
        "--find-links",
        find_links[0],
        *TARGET,
        pkg,
    ]

    # As single string (just use the first element from find_links)
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, find_links=find_links[0])
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Invalid proto raises exception
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(
            CommandExecutionError,
            pip.install,
            "'" + pkg + "'",
            find_links="sftp://pypi.crate.io",
        )


def test_install_no_index_with_index_url_or_extra_index_url_raises():
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(
            CommandExecutionError,
            pip.install,
            no_index=True,
            index_url="http://foo.tld",
        )

    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(
            CommandExecutionError,
            pip.install,
            no_index=True,
            extra_index_url="http://foo.tld",
        )


def test_install_failed_cached_requirements():
    with patch("salt.modules.pip._get_cached_requirements") as get_cached_requirements:
        get_cached_requirements.return_value = False
        ret = pip.install(requirements="salt://my_test_reqs")
        assert False is ret["result"]
        assert "my_test_reqs" in ret["comment"]


def test_install_cached_requirements_used(python_binary):
    with patch("salt.modules.pip._get_cached_requirements") as get_cached_requirements:
        get_cached_requirements.return_value = "my_cached_reqs"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(requirements="salt://requirements.txt")
            expected = [
                *python_binary,
                "install",
                "--requirement",
                "my_cached_reqs",
                *TARGET,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_install_venv():
    with patch("os.path") as mock_path:

        def join(*args):
            return os.path.normpath(os.sep.join(args))

        mock_path.is_file.return_value = True
        mock_path.isdir.return_value = True
        mock_path.join = join

        if salt.utils.platform.is_windows():
            venv_path = "C:\\test_env"
            bin_path = os.path.join(venv_path, "python.exe")
        else:
            venv_path = "/test_env"
            bin_path = os.path.join(venv_path, "python")

        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        pip_bin = MagicMock(return_value=[bin_path, "-m", "pip"])

        with patch.dict(pip.__salt__, {"cmd.run_all": mock}), patch.object(
            pip, "_get_pip_bin", pip_bin
        ):
            pip.install("mock", bin_env=venv_path)
            mock.assert_called_with(
                [bin_path, "-m", "pip", "install", "mock"],
                env={"VIRTUAL_ENV": venv_path},
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_install_log_argument_in_resulting_command(python_binary, tmp_path):
    with patch("os.access") as mock_path:
        pkg = "pep8"
        log_path = str(tmp_path / "pip-install.log")
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, log=log_path)
            expected = [
                *python_binary,
                "install",
                "--log",
                log_path,
                *TARGET,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_non_writeable_log():
    with patch("os.path") as mock_path:
        # Let's fake a non-writable log file
        pkg = "pep8"
        log_path = "/tmp/pip-install.log"
        mock_path.exists.side_effect = IOError("Fooo!")
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pytest.raises(IOError, pip.install, pkg, log=log_path)


def test_install_timeout_argument_in_resulting_command(python_binary):
    # Passing an int
    pkg = "pep8"
    expected = [*python_binary, "install", "--timeout"]
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, timeout=10)
        mock.assert_called_with(
            expected + [10, *TARGET, pkg],
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing an int as a string
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, timeout="10")
        mock.assert_called_with(
            expected + ["10", *TARGET, pkg],
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing a non-int to timeout
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(ValueError, pip.install, pkg, timeout="a")


def test_install_index_url_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    index_url = "http://foo.tld"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, index_url=index_url)
        expected = [
            *python_binary,
            "install",
            "--index-url",
            index_url,
            *TARGET,
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_extra_index_url_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    extra_index_url = "http://foo.tld"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, extra_index_url=extra_index_url)
        expected = [
            *python_binary,
            "install",
            "--extra-index-url",
            extra_index_url,
            *TARGET,
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_no_index_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, no_index=True)
        expected = [*python_binary, "install", "--no-index", *TARGET, pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_build_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    build = "/tmp/foo"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, build=build)
        expected = [*python_binary, "install", "--build", build, *TARGET, pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_target_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    target = "/tmp/foo"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, target=target)
        expected = [*python_binary, "install", "--target", target, pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_download_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    download = "/tmp/foo"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, download=download)
        expected = [
            *python_binary,
            "install",
            *TARGET,
            "--download",
            download,
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_no_download_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, no_download=True)
        expected = [*python_binary, "install", *TARGET, "--no-download", pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_download_cache_dir_arguments_in_resulting_command(python_binary):
    pkg = "pep8"
    cache_dir_arg_mapping = {
        "1.5.6": "--download-cache",
        "6.0": "--cache-dir",
    }
    download_cache = "/tmp/foo"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})

    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        for pip_version, cmd_arg in cache_dir_arg_mapping.items():
            with patch("salt.modules.pip.version", MagicMock(return_value=pip_version)):
                # test `download_cache` kwarg
                pip.install(pkg, download_cache="/tmp/foo")
                expected = [
                    *python_binary,
                    "install",
                    *TARGET,
                    cmd_arg,
                    download_cache,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

                # test `cache_dir` kwarg
                pip.install(pkg, cache_dir="/tmp/foo")
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )


def test_install_source_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    source = "/tmp/foo"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, source=source)
        expected = [*python_binary, "install", *TARGET, "--source", source, pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_exists_action_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    for action in ("s", "i", "w", "b"):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, exists_action=action)
            expected = [
                *python_binary,
                "install",
                *TARGET,
                "--exists-action",
                action,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    # Test for invalid action
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(CommandExecutionError, pip.install, pkg, exists_action="d")


def test_install_install_options_argument_in_resulting_command(python_binary):
    install_options = ["--exec-prefix=/foo/bar", "--install-scripts=/foo/bar/bin"]
    pkg = "pep8"

    expected = [*python_binary, "install", *TARGET]
    for item in install_options:
        expected.extend(["--install-option", item])
    expected.append(pkg)

    # Passing options as a list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, install_options=install_options)
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing mirrors as a comma separated list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, install_options=",".join(install_options))
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing mirrors as a single string entry
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, install_options=install_options[0])
        expected = [
            *python_binary,
            "install",
            *TARGET,
            "--install-option",
            install_options[0],
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_global_options_argument_in_resulting_command(python_binary):
    global_options = ["--quiet", "--no-user-cfg"]
    pkg = "pep8"

    expected = [*python_binary, "install", *TARGET]
    for item in global_options:
        expected.extend(["--global-option", item])
    expected.append(pkg)

    # Passing options as a list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, global_options=global_options)
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing mirrors as a comma separated list
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, global_options=",".join(global_options))
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing mirrors as a single string entry
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, global_options=global_options[0])
        expected = [
            *python_binary,
            "install",
            *TARGET,
            "--global-option",
            global_options[0],
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_upgrade_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, upgrade=True)
        expected = [*python_binary, "install", *TARGET, "--upgrade", pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_force_reinstall_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, force_reinstall=True)
        expected = [
            *python_binary,
            "install",
            *TARGET,
            "--force-reinstall",
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_ignore_installed_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, ignore_installed=True)
        expected = [
            *python_binary,
            "install",
            *TARGET,
            "--ignore-installed",
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_no_deps_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, no_deps=True)
        expected = [*python_binary, "install", *TARGET, "--no-deps", pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_no_install_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, no_install=True)
        expected = [*python_binary, "install", *TARGET, "--no-install", pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_proxy_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    proxy = "salt-user:salt-passwd@salt-proxy:3128"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(pkg, proxy=proxy)
        expected = [*python_binary, "install", "--proxy", proxy, *TARGET, pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_proxy_false_argument_in_resulting_command(python_binary):
    """
    Checking that there is no proxy set if proxy arg is set to False
    even if the global proxy is set.
    """
    pkg = "pep8"
    proxy = False
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    config_mock = {
        "proxy_host": "salt-proxy",
        "proxy_port": "3128",
        "proxy_username": "salt-user",
        "proxy_password": "salt-passwd",
    }
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch.dict(pip.__opts__, config_mock):
            pip.install(pkg, proxy=proxy)
            expected = [*python_binary, "install", *TARGET, pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_install_global_proxy_in_resulting_command(python_binary):
    """
    Checking that there is proxy set if global proxy is set.
    """
    pkg = "pep8"
    proxy = "http://salt-user:salt-passwd@salt-proxy:3128"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    config_mock = {
        "proxy_host": "salt-proxy",
        "proxy_port": "3128",
        "proxy_username": "salt-user",
        "proxy_password": "salt-passwd",
    }
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch.dict(pip.__opts__, config_mock):
            pip.install(pkg)
            expected = [
                *python_binary,
                "install",
                "--proxy",
                proxy,
                *TARGET,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_install_multiple_requirements_arguments_in_resulting_command(python_binary):
    with patch("salt.modules.pip._get_cached_requirements") as get_cached_requirements:
        cached_reqs = ["my_cached_reqs-1", "my_cached_reqs-2"]
        get_cached_requirements.side_effect = cached_reqs
        requirements = ["salt://requirements-1.txt", "salt://requirements-2.txt"]

        expected = [*python_binary, "install"]
        for item in cached_reqs:
            expected.extend(["--requirement", item])
        expected.extend(TARGET)

        # Passing option as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(requirements=requirements)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing option as a comma separated list
        get_cached_requirements.side_effect = cached_reqs
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(requirements=",".join(requirements))
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing option as a single string entry
        get_cached_requirements.side_effect = [cached_reqs[0]]
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(requirements=requirements[0])
            expected = [
                *python_binary,
                "install",
                "--requirement",
                cached_reqs[0],
                *TARGET,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_install_extra_args_arguments_in_resulting_command(python_binary):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.install(
            pkg, extra_args=[{"--latest-pip-kwarg": "param"}, "--latest-pip-arg"]
        )
        expected = [
            *python_binary,
            "install",
            *TARGET,
            pkg,
            "--latest-pip-kwarg",
            "param",
            "--latest-pip-arg",
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_install_extra_args_arguments_recursion_error():
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(
            TypeError,
            lambda: pip.install(
                pkg, extra_args=[{"--latest-pip-kwarg": ["param1", "param2"]}]
            ),
        )

        pytest.raises(
            TypeError,
            lambda: pip.install(
                pkg, extra_args=[{"--latest-pip-kwarg": [{"--too-deep": dict()}]}]
            ),
        )


def test_uninstall_multiple_requirements_arguments_in_resulting_command(python_binary):
    with patch("salt.modules.pip._get_cached_requirements") as get_cached_requirements:
        cached_reqs = ["my_cached_reqs-1", "my_cached_reqs-2"]
        get_cached_requirements.side_effect = cached_reqs
        requirements = ["salt://requirements-1.txt", "salt://requirements-2.txt"]

        expected = [*python_binary, "uninstall", "-y"]
        for item in cached_reqs:
            expected.extend(["--requirement", item])

        # Passing option as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(requirements=requirements)
            mock.assert_called_with(
                expected,
                cwd=None,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing option as a comma separated list
        get_cached_requirements.side_effect = cached_reqs
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(requirements=",".join(requirements))
            mock.assert_called_with(
                expected,
                cwd=None,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing option as a single string entry
        get_cached_requirements.side_effect = [cached_reqs[0]]
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(requirements=requirements[0])
            expected = [
                *python_binary,
                "uninstall",
                "-y",
                "--requirement",
                cached_reqs[0],
            ]
            mock.assert_called_with(
                expected,
                cwd=None,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_uninstall_global_proxy_in_resulting_command(python_binary):
    """
    Checking that there is proxy set if global proxy is set.
    """
    pkg = "pep8"
    proxy = "http://salt-user:salt-passwd@salt-proxy:3128"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    config_mock = {
        "proxy_host": "salt-proxy",
        "proxy_port": "3128",
        "proxy_username": "salt-user",
        "proxy_password": "salt-passwd",
    }
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch.dict(pip.__opts__, config_mock):
            pip.uninstall(pkg)
            expected = [
                *python_binary,
                "uninstall",
                "-y",
                "--proxy",
                proxy,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                cwd=None,
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_uninstall_proxy_false_argument_in_resulting_command(python_binary):
    """
    Checking that there is no proxy set if proxy arg is set to False
    even if the global proxy is set.
    """
    pkg = "pep8"
    proxy = False
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    config_mock = {
        "proxy_host": "salt-proxy",
        "proxy_port": "3128",
        "proxy_username": "salt-user",
        "proxy_password": "salt-passwd",
    }
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch.dict(pip.__opts__, config_mock):
            pip.uninstall(pkg, proxy=proxy)
            expected = [*python_binary, "uninstall", "-y", pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                cwd=None,
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_uninstall_log_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    log_path = "/tmp/pip-install.log"

    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.uninstall(pkg, log=log_path)
        expected = [
            *python_binary,
            "uninstall",
            "-y",
            "--log",
            log_path,
            pkg,
        ]
        mock.assert_called_with(
            expected,
            saltenv="base",
            cwd=None,
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Let's fake a non-writable log file
    with patch("os.path") as mock_path:
        mock_path.exists.side_effect = IOError("Fooo!")
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pytest.raises(IOError, pip.uninstall, pkg, log=log_path)


def test_uninstall_timeout_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    expected = [*python_binary, "uninstall", "-y", "--timeout"]
    # Passing an int
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.uninstall(pkg, timeout=10)
        mock.assert_called_with(
            expected + [10, pkg],
            cwd=None,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing an int as a string
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pip.uninstall(pkg, timeout="10")
        mock.assert_called_with(
            expected + ["10", pkg],
            cwd=None,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )

    # Passing a non-int to timeout
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        pytest.raises(ValueError, pip.uninstall, pkg, timeout="a")


def test_freeze_command(python_binary):
    expected = [*python_binary, "freeze"]
    eggs = [
        "M2Crypto==0.21.1",
        "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
        "bbfreeze==1.1.0",
        "bbfreeze-loader==1.1.0",
        "pycrypto==2.6",
    ]
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            ret = pip.freeze()
            mock.assert_called_with(
                expected,
                cwd=None,
                runas=None,
                use_vt=False,
                python_shell=False,
            )
            assert ret == eggs

    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    # Passing env_vars passes them to underlying command?
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            ret = pip.freeze(env_vars={"foo": "bar"})
            mock.assert_called_with(
                expected,
                cwd=None,
                runas=None,
                use_vt=False,
                python_shell=False,
                env={"foo": "bar"},
            )
            assert ret == eggs

    # Non zero returncode raises exception?
    mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            pytest.raises(
                CommandExecutionError,
                pip.freeze,
            )


def test_freeze_command_with_all(python_binary):
    eggs = [
        "M2Crypto==0.21.1",
        "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
        "bbfreeze==1.1.0",
        "bbfreeze-loader==1.1.0",
        "pip==0.9.1",
        "pycrypto==2.6",
        "setuptools==20.10.1",
    ]
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="9.0.1")):
            ret = pip.freeze()
            expected = [*python_binary, "freeze", "--all"]
            mock.assert_called_with(
                expected,
                cwd=None,
                runas=None,
                use_vt=False,
                python_shell=False,
            )
            assert ret == eggs

    # Non zero returncode raises exception?
    mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="9.0.1")):
            pytest.raises(
                CommandExecutionError,
                pip.freeze,
            )


def test_list_freeze_parse_command(python_binary):
    eggs = [
        "M2Crypto==0.21.1",
        "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
        "bbfreeze==1.1.0",
        "bbfreeze-loader==1.1.0",
        "pycrypto==2.6",
    ]
    mock_version = "6.1.1"
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value=mock_version)):
            ret = pip.list_freeze_parse()
            expected = [*python_binary, "freeze"]
            mock.assert_called_with(
                expected,
                cwd=None,
                runas=None,
                python_shell=False,
                use_vt=False,
            )
            assert ret == {
                "SaltTesting-dev": "git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8",
                "M2Crypto": "0.21.1",
                "bbfreeze-loader": "1.1.0",
                "bbfreeze": "1.1.0",
                "pip": mock_version,
                "pycrypto": "2.6",
            }

    # Non zero returncode raises exception?
    mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            pytest.raises(
                CommandExecutionError,
                pip.list_freeze_parse,
            )


def test_list_freeze_parse_command_with_all(python_binary):
    eggs = [
        "M2Crypto==0.21.1",
        "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
        "bbfreeze==1.1.0",
        "bbfreeze-loader==1.1.0",
        "pip==9.0.1",
        "pycrypto==2.6",
        "setuptools==20.10.1",
    ]
    # N.B.: this is deliberately different from the "output" of pip freeze.
    # This is to demonstrate that the version reported comes from freeze
    # instead of from the pip.version function.
    mock_version = "9.0.0"
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value=mock_version)):
            ret = pip.list_freeze_parse()
            expected = [*python_binary, "freeze", "--all"]
            mock.assert_called_with(
                expected,
                cwd=None,
                runas=None,
                python_shell=False,
                use_vt=False,
            )
            assert ret == {
                "SaltTesting-dev": "git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8",
                "M2Crypto": "0.21.1",
                "bbfreeze-loader": "1.1.0",
                "bbfreeze": "1.1.0",
                "pip": "9.0.1",
                "pycrypto": "2.6",
                "setuptools": "20.10.1",
            }

    # Non zero returncode raises exception?
    mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            pytest.raises(
                CommandExecutionError,
                pip.list_freeze_parse,
            )


def test_list_freeze_parse_command_with_prefix(python_binary):
    eggs = [
        "M2Crypto==0.21.1",
        "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
        "bbfreeze==1.1.0",
        "bbfreeze-loader==1.1.0",
        "pycrypto==2.6",
    ]
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            ret = pip.list_freeze_parse(prefix="bb")
            expected = [*python_binary, "freeze"]
            mock.assert_called_with(
                expected,
                cwd=None,
                runas=None,
                python_shell=False,
                use_vt=False,
            )
            assert ret == {"bbfreeze-loader": "1.1.0", "bbfreeze": "1.1.0"}


def test_list_upgrades_legacy(python_binary):
    eggs = [
        "apache-libcloud (Current: 1.1.0 Latest: 2.2.1 [wheel])",
        "appdirs (Current: 1.4.1 Latest: 1.4.3 [wheel])",
        "awscli (Current: 1.11.63 Latest: 1.12.1 [sdist])",
    ]
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            ret = pip.list_upgrades()
            mock.assert_called_with(
                [*python_binary, "list", "--outdated"],
                cwd=None,
                runas=None,
            )
            assert ret == {
                "apache-libcloud": "2.2.1 [wheel]",
                "appdirs": "1.4.3 [wheel]",
                "awscli": "1.12.1 [sdist]",
            }


def test_list_upgrades_gt9(python_binary):
    eggs = """[{"latest_filetype": "wheel", "version": "1.1.0", "name": "apache-libcloud", "latest_version": "2.2.1"},
            {"latest_filetype": "wheel", "version": "1.4.1", "name": "appdirs", "latest_version": "1.4.3"},
            {"latest_filetype": "sdist", "version": "1.11.63", "name": "awscli", "latest_version": "1.12.1"}
            ]"""
    mock = MagicMock(return_value={"retcode": 0, "stdout": f"{eggs}"})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="9.1.1")):
            ret = pip.list_upgrades()
            mock.assert_called_with(
                [
                    *python_binary,
                    "list",
                    "--outdated",
                    "--format=json",
                ],
                cwd=None,
                runas=None,
            )
            assert ret == {
                "apache-libcloud": "2.2.1 [wheel]",
                "appdirs": "1.4.3 [wheel]",
                "awscli": "1.12.1 [sdist]",
            }


def test_is_installed_true(python_binary):
    eggs = [
        "M2Crypto==0.21.1",
        "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
        "bbfreeze==1.1.0",
        "bbfreeze-loader==1.1.0",
        "pycrypto==2.6",
    ]
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            ret = pip.is_installed(pkgname="bbfreeze")
            mock.assert_called_with(
                [*python_binary, "freeze"],
                cwd=None,
                runas=None,
                python_shell=False,
                use_vt=False,
            )
            assert ret


def test_is_installed_false(python_binary):
    eggs = [
        "M2Crypto==0.21.1",
        "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
        "bbfreeze==1.1.0",
        "bbfreeze-loader==1.1.0",
        "pycrypto==2.6",
    ]
    mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
            ret = pip.is_installed(pkgname="notexist")
            mock.assert_called_with(
                [*python_binary, "freeze"],
                cwd=None,
                runas=None,
                python_shell=False,
                use_vt=False,
            )
            assert not ret


def test_install_pre_argument_in_resulting_command(python_binary):
    pkg = "pep8"
    # Lower than 1.4 versions don't end up with `--pre` in the resulting output
    mock = MagicMock(
        side_effect=[
            {"retcode": 0, "stdout": "pip 1.2.0 /path/to/site-packages/pip"},
            {"retcode": 0, "stdout": ""},
        ]
    )
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="1.3")):
            pip.install(pkg, pre_releases=True)
            expected = [*python_binary, "install", *TARGET, pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    mock_run = MagicMock(return_value="pip 1.4.1 /path/to/site-packages/pip")
    mock_run_all = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(
        pip.__salt__, {"cmd.run_stdout": mock_run, "cmd.run_all": mock_run_all}
    ):
        with patch("salt.modules.pip._get_pip_bin", MagicMock(return_value=["pip"])):
            pip.install(pkg, pre_releases=True)
            expected = ["pip", "install", *TARGET, "--pre", pkg]
            mock_run_all.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )


def test_resolve_requirements_chain_function():
    with patch("salt.utils.files.fopen", FakeFopen):
        chain = pip._resolve_requirements_chain(
            ["requirements-0.txt", "requirements-3.txt"]
        )
    assert chain == [
        "requirements-0.txt",
        "requirements-1.txt",
        "requirements-2.txt",
        "requirements-3.txt",
        "requirements-4.txt",
    ]


def test_when_upgrade_is_called_and_there_are_available_upgrades_it_should_call_correct_command(
    expected_user,
):
    fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": "{}"})
    pip_user = expected_user
    with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
        "salt.modules.pip.list_upgrades", autospec=True, return_value=[pip_user]
    ), patch(
        "salt.modules.pip._get_pip_bin",
        autospec=True,
        return_value=["some-other-pip"],
    ):
        pip.upgrade(user=pip_user)

        fake_run_all.assert_any_call(
            ["some-other-pip", "install", "-U", "list", "--format=json", pip_user],
            runas=pip_user,
            cwd=None,
            use_vt=False,
        )


def test_when_list_upgrades_is_provided_a_user_it_should_be_passed_to_the_version_command(
    expected_user,
):
    fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": "{}"})
    pip_user = expected_user

    def all_new_commands(*args, **kwargs):
        """
        Without this, mutating the return value mutates the return value
        for EVERYTHING.
        """
        return ["some-other-pip"]

    with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
        "salt.modules.pip._get_pip_bin",
        autospec=True,
        side_effect=all_new_commands,
    ):
        pip._clear_context()
        pip.list_upgrades(user=pip_user)
        fake_run_all.assert_any_call(
            ["some-other-pip", "--version"],
            runas=expected_user,
            cwd=None,
            python_shell=False,
        )


def test_when_install_is_provided_a_user_it_should_be_passed_to_the_version_command(
    expected_user,
):
    fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": "{}"})
    pip_user = expected_user

    def all_new_commands(*args, **kwargs):
        """
        Without this, mutating the return value mutates the return value
        for EVERYTHING.
        """
        return ["some-other-pip"]

    with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
        "salt.modules.pip._get_pip_bin",
        autospec=True,
        side_effect=all_new_commands,
    ):
        pip._clear_context()
        pip.install(user=pip_user)
        fake_run_all.assert_any_call(
            ["some-other-pip", "--version"],
            runas=pip_user,
            cwd=None,
            python_shell=False,
        )


def test_when_version_is_called_with_a_user_it_should_be_passed_to_undelying_runas(
    expected_user,
):
    fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": ""})
    pip_user = expected_user
    with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
        "salt.modules.pip.list_upgrades", autospec=True, return_value=[pip_user]
    ), patch(
        "salt.modules.pip._get_pip_bin",
        autospec=True,
        return_value=["some-new-pip"],
    ):
        pip.version(user=pip_user)
        fake_run_all.assert_called_with(
            ["some-new-pip", "--version"],
            runas=pip_user,
            cwd=None,
            python_shell=False,
        )


@pytest.mark.parametrize(
    "bin_env,target,target_env,expected_target",
    [
        (None, None, None, None),
        (None, "/tmp/foo", None, "/tmp/foo"),
        (None, None, "/tmp/bar", "/tmp/bar"),
        (None, "/tmp/foo", "/tmp/bar", "/tmp/foo"),
        ("/tmp/venv", "/tmp/foo", None, "/tmp/foo"),
        ("/tmp/venv", None, "/tmp/bar", None),
        ("/tmp/venv", "/tmp/foo", "/tmp/bar", "/tmp/foo"),
    ],
)
def test_install_target_from_VENV_PIP_TARGET_in_resulting_command(
    python_binary, bin_env, target, target_env, expected_target
):
    pkg = "pep8"
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    environment = os.environ.copy()
    real_get_pip_bin = pip._get_pip_bin

    def mock_get_pip_bin(bin_env):
        if not bin_env:
            return real_get_pip_bin(bin_env)
        return [f"{bin_env}/bin/pip"]

    if target_env is not None:
        environment["VENV_PIP_TARGET"] = target_env
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}), patch.object(
        os, "environ", environment
    ), patch.object(pip, "_get_pip_bin", mock_get_pip_bin):
        pip.install(pkg, bin_env=bin_env, target=target)
        expected_binary = python_binary
        if bin_env is not None:
            expected_binary = [f"{bin_env}/bin/pip"]
        if expected_target is not None:
            expected = [*expected_binary, "install", "--target", expected_target, pkg]
        else:
            expected = [*expected_binary, "install", pkg]
        mock.assert_called_with(
            expected,
            saltenv="base",
            runas=None,
            use_vt=False,
            python_shell=False,
        )


def test_list(python_binary):
    json_out = dedent(
        """
    [
      {
        "name": "idemenv",
        "version": "0.2.0",
        "editable_project_location": "/home/debian/idemenv"
      },
      {
        "name": "MarkupSafe",
        "version": "2.1.1"
      },
      {
        "name": "pip",
        "version": "22.3.1"
      },
      {
        "name": "pop",
        "version": "23.0.0"
      },
      {
        "name": "salt",
        "version": "3006.0+0na.5b18e86"
      },
      {
        "name": "typing_extensions",
        "version": "4.4.0"
      },
      {
        "name": "unattended-upgrades",
        "version": "0.1"
      },
      {
        "name": "yarl",
        "version": "1.8.2"
      }
    ]
    """
    )
    mock_version = "22.3.1"
    mock = MagicMock(return_value={"retcode": 0, "stdout": json_out})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value=mock_version)):
            ret = pip.list_()
            expected = [*python_binary, "list", "--format=json"]
            mock.assert_called_with(
                expected,
                cwd=None,
                runas=None,
                python_shell=False,
            )
            assert ret == {
                "MarkupSafe": "2.1.1",
                "idemenv": "0.2.0",
                "pip": "22.3.1",
                "pop": "23.0.0",
                "salt": "3006.0+0na.5b18e86",
                "typing_extensions": "4.4.0",
                "unattended-upgrades": "0.1",
                "yarl": "1.8.2",
            }

    # Non zero returncode raises exception?
    mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
    with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
        with patch("salt.modules.pip.version", MagicMock(return_value="22.3.1")):
            pytest.raises(
                CommandExecutionError,
                pip.list_,
            )



"""
    TestCase for the salt.modules.vagrant module.
"""

import pytest

import salt.exceptions
import salt.modules.vagrant as vagrant
import salt.utils.platform
from tests.support.mock import MagicMock, patch


@pytest.fixture
def local_opts(tmp_path):
    return {
        "extension_modules": "",
        "vagrant_sdb_data": {
            "driver": "sqlite3",
            "database": str(tmp_path / "test_vagrant.sqlite"),
            "table": "sdb",
            "create_table": True,
        },
    }


@pytest.fixture
def configure_loader_modules(local_opts):
    return {vagrant: {"__opts__": local_opts}}


def test_vagrant_get_vm_info_not_found():
    mock_sdb = MagicMock(return_value=None)
    with patch.dict(vagrant.__utils__, {"sdb.sdb_get": mock_sdb}):
        with pytest.raises(salt.exceptions.SaltInvocationError):
            vagrant.get_vm_info("thisNameDoesNotExist")


def test_vagrant_init_positional(local_opts, tmp_path):
    path_nowhere = str(tmp_path / "tmp" / "nowhere")
    mock_sdb = MagicMock(return_value=None)
    with patch.dict(vagrant.__utils__, {"sdb.sdb_set": mock_sdb}):
        resp = vagrant.init(
            "test1",
            path_nowhere,
            "onetest",
            "nobody",
            False,
            "french",
            {"different": "very"},
        )
        assert resp.startswith("Name test1 defined")
        expected = dict(
            name="test1",
            cwd=path_nowhere,
            machine="onetest",
            runas="nobody",
            vagrant_provider="french",
            different="very",
        )
        mock_sdb.assert_called_with(
            f"sdb://vagrant_sdb_data/onetest?{path_nowhere}",
            "test1",
            local_opts,
        )
        mock_sdb.assert_any_call("sdb://vagrant_sdb_data/test1", expected, local_opts)


def test_vagrant_get_vm_info():
    testdict = {"testone": "one", "machine": "two"}
    mock_sdb = MagicMock(return_value=testdict)
    with patch.dict(vagrant.__utils__, {"sdb.sdb_get": mock_sdb}):
        resp = vagrant.get_vm_info("test1")
        assert resp == testdict


def test_vagrant_init_dict(local_opts):
    testdict = dict(
        cwd="/tmp/anywhere",
        machine="twotest",
        runas="somebody",
        vagrant_provider="english",
    )
    expected = testdict.copy()
    expected["name"] = "test2"
    mock_sdb = MagicMock(return_value=None)
    with patch.dict(vagrant.__utils__, {"sdb.sdb_set": mock_sdb}):
        vagrant.init("test2", vm=testdict)
        mock_sdb.assert_any_call("sdb://vagrant_sdb_data/test2", expected, local_opts)


def test_vagrant_init_arg_override(local_opts):
    testdict = dict(
        cwd="/tmp/there",
        machine="treetest",
        runas="anybody",
        vagrant_provider="spansh",
    )
    mock_sdb = MagicMock(return_value=None)
    with patch.dict(vagrant.__utils__, {"sdb.sdb_set": mock_sdb}):
        vagrant.init(
            "test3",
            cwd="/tmp",
            machine="threetest",
            runas="him",
            vagrant_provider="polish",
            vm=testdict,
        )
        expected = dict(
            name="test3",
            cwd="/tmp",
            machine="threetest",
            runas="him",
            vagrant_provider="polish",
        )
        mock_sdb.assert_any_call("sdb://vagrant_sdb_data/test3", expected, local_opts)


def test_vagrant_get_ssh_config_fails():
    mock_sdb = MagicMock(return_value=None)
    with patch.dict(vagrant.__utils__, {"sdb.sdb_set": mock_sdb}):
        mock_sdb = MagicMock(return_value={})
        with patch.dict(vagrant.__utils__, {"sdb.sdb_get": mock_sdb}):
            vagrant.init("test3", cwd="/tmp")
            with pytest.raises(salt.exceptions.SaltInvocationError):
                vagrant.get_ssh_config("test3")  # has not been started


def test_vagrant_destroy(local_opts, tmp_path):
    path_mydir = str(tmp_path / "my" / "dir")
    mock_cmd = MagicMock(return_value={"retcode": 0})
    with patch.dict(vagrant.__salt__, {"cmd.run_all": mock_cmd}):
        mock_sdb = MagicMock(return_value=None)
        with patch.dict(vagrant.__utils__, {"sdb.sdb_delete": mock_sdb}):
            mock_sdb_get = MagicMock(
                return_value={"machine": "macfour", "cwd": path_mydir}
            )
            with patch.dict(vagrant.__utils__, {"sdb.sdb_get": mock_sdb_get}):
                assert vagrant.destroy("test4")
                mock_sdb.assert_any_call(
                    f"sdb://vagrant_sdb_data/macfour?{path_mydir}",
                    local_opts,
                )
                mock_sdb.assert_any_call("sdb://vagrant_sdb_data/test4", local_opts)
                cmd = "vagrant destroy -f macfour"
                mock_cmd.assert_called_with(
                    cmd, runas=None, cwd=path_mydir, output_loglevel="info"
                )


def test_vagrant_start():
    mock_cmd = MagicMock(return_value={"retcode": 0})
    with patch.dict(vagrant.__salt__, {"cmd.run_all": mock_cmd}):
        mock_sdb_get = MagicMock(
            return_value={
                "machine": "five",
                "cwd": "/the/dir",
                "runas": "me",
                "vagrant_provider": "him",
            }
        )
        with patch.dict(vagrant.__utils__, {"sdb.sdb_get": mock_sdb_get}):
            assert vagrant.start("test5")
            cmd = "vagrant up five --provider=him"
            mock_cmd.assert_called_with(
                cmd, runas="me", cwd="/the/dir", output_loglevel="info"
            )



import logging

import pytest

import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
import salt.utils.win_dacl as win_dacl
import salt.utils.win_lgpo_auditpol as auditpol
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cachedir = tmp_path / "__test_admx_policy_cache_dir"
    cachedir.mkdir(parents=True, exist_ok=True)
    return {
        win_lgpo: {
            "__opts__": {"cachedir": cachedir},
            "__salt__": {
                "file.copy": win_file.copy,
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
                "file.remove": win_file.remove,
                "file.write": win_file.write,
            },
            "__utils__": {
                "auditpol.get_auditpol_dump": auditpol.get_auditpol_dump,
                "auditpol.set_setting": auditpol.set_setting,
            },
        },
        auditpol: {
            "__context__": {},
        },
        win_file: {
            "__utils__": {
                "dacl.set_perms": win_dacl.set_perms,
            },
        },
    }


@pytest.fixture(scope="module")
def disable_legacy_auditing():
    # To test and use these policy settings we have to set one of the policies to Enabled
    # Location: Windows Settings -> Security Settings -> Local Policies -> Security Options
    # Policy: "Audit: Force audit policy subcategory settings..."
    # Short Name: SceNoApplyLegacyAuditPolicy
    from tests.support.sminion import create_sminion

    salt_minion = create_sminion()
    test_setting = "Enabled"
    pre_security_setting = salt_minion.functions.lgpo.get_policy(
        policy_name="SceNoApplyLegacyAuditPolicy", policy_class="machine"
    )
    pre_audit_setting = salt_minion.functions.lgpo.get_policy(
        policy_name="Audit User Account Management", policy_class="machine"
    )
    try:
        if pre_security_setting != test_setting:
            salt_minion.functions.lgpo.set_computer_policy(
                name="SceNoApplyLegacyAuditPolicy", setting=test_setting
            )
            assert (
                salt_minion.functions.lgpo.get_policy(
                    policy_name="SceNoApplyLegacyAuditPolicy", policy_class="machine"
                )
                == test_setting
            )
        yield
    finally:
        salt_minion.functions.lgpo.set_computer_policy(
            name="SceNoApplyLegacyAuditPolicy", setting=pre_security_setting
        )
        salt_minion.functions.lgpo.set_computer_policy(
            name="Audit User Account Management", setting=pre_audit_setting
        )


@pytest.fixture
def clear_policy():
    # Ensure the policy is not set
    test_setting = "No Auditing"
    win_lgpo.set_computer_policy(
        name="Audit User Account Management", setting=test_setting
    )
    assert (
        win_lgpo.get_policy(
            policy_name="Audit User Account Management", policy_class="machine"
        )
        == test_setting
    )


@pytest.fixture
def set_policy():
    # Ensure the policy is set
    test_setting = "Success"
    win_lgpo.set_computer_policy(
        name="Audit User Account Management", setting=test_setting
    )
    assert (
        win_lgpo.get_policy(
            policy_name="Audit User Account Management", policy_class="machine"
        )
        == test_setting
    )


@pytest.mark.parametrize(
    "setting, expected",
    [
        ("No Auditing", "0"),
        ("Success", "1"),
        ("Failure", "2"),
        ("Success and Failure", "3"),
    ],
)
def test_get_value(setting, expected):
    """
    Helper function to set an audit setting and assert that it was successful
    """
    win_lgpo.set_computer_policy(name="Audit User Account Management", setting=setting)
    # Clear the context so we're getting the actual settings from the machine
    result = win_lgpo._get_advaudit_value("Audit User Account Management", refresh=True)
    assert result == expected


def test_get_defaults():
    patch_context = patch.dict(win_lgpo.__context__, {})
    patch_salt = patch.dict(
        win_lgpo.__utils__, {"auditpol.get_auditpol_dump": auditpol.get_auditpol_dump}
    )
    with patch_context, patch_salt:
        assert "Machine Name" in win_lgpo._get_advaudit_defaults("fieldnames")

    audit_defaults = {"junk": "defaults"}
    patch_context = patch.dict(
        win_lgpo.__context__, {"lgpo.audit_defaults": audit_defaults}
    )
    with patch_context, patch_salt:
        assert win_lgpo._get_advaudit_defaults() == audit_defaults


def test_set_value_error():
    mock_set_file_data = MagicMock(return_value=False)
    with patch.object(win_lgpo, "_set_advaudit_file_data", mock_set_file_data):
        with pytest.raises(CommandExecutionError):
            win_lgpo._set_advaudit_value("Audit User Account Management", "None")


def test_set_value_log_messages(caplog):
    mock_set_file_data = MagicMock(return_value=True)
    mock_set_pol_data = MagicMock(return_value=False)
    mock_context = {"lgpo.adv_audit_data": {"test_option": "test_value"}}
    with caplog.at_level(logging.DEBUG):
        with patch.object(
            win_lgpo, "_set_advaudit_file_data", mock_set_file_data
        ), patch.object(
            win_lgpo, "_set_advaudit_pol_data", mock_set_pol_data
        ), patch.dict(
            win_lgpo.__context__, mock_context
        ):
            win_lgpo._set_advaudit_value("test_option", None)
            assert "Failed to apply audit setting:" in caplog.text
        assert "LGPO: Removing Advanced Audit data:" in caplog.text



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest
from saltfactories.utils import random_string

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def group(modules):
    return modules.group


# Create group name strings for tests
@pytest.fixture(scope="function")
def non_existing_group_name(group):
    group_name = random_string("group-", lowercase=False)
    try:
        yield group_name
    finally:
        # Delete the added group
        group_info = group.info(group_name)
        if group_info:
            group.delete(group_name)


@pytest.fixture(scope="function")
def existing_group_name(group):
    group_name = random_string("group-", lowercase=False)
    try:
        ret = group.add(group_name, 4567)
        if ret is not True:
            pytest.skip("Failed to create a group to delete")
        yield group_name
    finally:
        # Delete the added group
        group_info = group.info(group_name)
        if group_info:
            group.delete(group_name)


@pytest.fixture(scope="function")
def non_existing_user(group):
    group_name = random_string("user-", lowercase=False)
    try:
        yield group_name
    finally:
        # Delete the added group
        group_info = group.info(group_name)
        if group_info:
            group.delete(group_name)


@pytest.fixture(scope="function")
def existing_user(group, existing_group_name):
    group_name = random_string("user-", lowercase=False)
    try:
        ret = group.adduser(existing_group_name, group_name)
        if ret is not True:
            pytest.skip("Failed to create an existing group member")
        yield group_name
    finally:
        # Delete the added group
        group_info = group.info(group_name)
        if group_info:
            group.delete(group_name)


@pytest.fixture(scope="module")
def rep_user_group():
    yield random_string("RS-", lowercase=False)


@pytest.fixture(scope="function")
def non_existing_group_member(group):
    group_name = random_string("user-", lowercase=False)
    try:
        yield group_name
    finally:
        # Delete the added group
        group_info = group.info(group_name)
        if group_info:
            group.delete(group_name)


def test_mac_group_add(group, non_existing_group_name):
    """
    Tests the add group function
    """
    group.add(non_existing_group_name, 3456)
    group_info = group.info(non_existing_group_name)
    assert group_info["name"] == non_existing_group_name


def test_mac_group_delete(group, existing_group_name):
    """
    Tests the delete group function
    """
    ret = group.delete(existing_group_name)
    assert ret


def test_mac_group_chgid(group, existing_group_name):
    """
    Tests changing the group id
    """
    gid = 6789
    group_info = group.info(existing_group_name)
    assert group_info["gid"] != gid
    group.chgid(existing_group_name, gid)
    group_info = group.info(existing_group_name)
    assert group_info["gid"] == gid


def test_mac_adduser(group, non_existing_group_name, non_existing_user):
    """
    Tests adding user to the group
    """
    # Create a group to use for test - If unsuccessful, skip the test
    ret = group.add(non_existing_group_name, 5678)
    if ret is not True:
        group.delete(non_existing_group_name)
        pytest.skip("Failed to create a group to manipulate")

    group.adduser(non_existing_group_name, non_existing_user)
    group_info = group.info(non_existing_group_name)
    assert non_existing_user in group_info["members"]
    assert group_info["members"] == [non_existing_user]


def test_mac_deluser(group, existing_group_name, existing_user):
    """
    Test deleting user from a group
    """
    delusr = group.deluser(existing_group_name, existing_user)
    assert delusr

    group_info = group.info(existing_group_name)
    assert existing_user not in group_info["members"]


def test_mac_members(
    group, existing_group_name, existing_user, non_existing_group_member
):
    """
    Test replacing members of a group
    """
    group.members(existing_group_name, existing_user)
    group_info = group.info(existing_group_name)
    assert non_existing_group_member not in group_info["members"]
    assert existing_user in group_info["members"]

    # Replace group members
    rep_group_mem = group.members(existing_group_name, non_existing_group_member)
    assert rep_group_mem

    # ensure new user is added to group and previous user is removed
    group_info = group.info(existing_group_name)
    assert non_existing_group_member in group_info["members"]
    assert existing_user not in group_info["members"]


def test_mac_getent(group, existing_user, existing_group_name):
    """
    Test returning info on all groups
    """
    getinfo = group.getent()
    assert getinfo
    assert existing_group_name in str(getinfo)
    assert existing_user in str(getinfo)



"""
Vistara Runner

Runner to interact with the Vistara (http://www.vistarait.com/) REST API

:codeauthor: Brad Thurber <brad.thurber@gmail.com>

To use this runner, the Vistara client_id and Vistara oauth2 client_key
and client_secret must be set in the master config.

For example ``/etc/salt/master.d/_vistara.conf``:

.. code-block:: yaml

    vistara:
      client_id: client_012345
      client_key: N0tReallyaR3alKeyButShouldB12345
      client_secret: ThisI5AreallyLongsecretKeyIwonderwhyTheyMakethemSoBigTheseDays00


"""

import logging

import salt.output

# See https://docs.saltproject.io/en/latest/topics/tutorials/http.html
import salt.utils.http

log = logging.getLogger(__name__)


def __virtual__():
    """
    Check to see if master config has the necessary config
    """
    vistara_config = __opts__["vistara"] if "vistara" in __opts__ else None

    if vistara_config:
        client_id = vistara_config.get("client_id", None)
        client_key = vistara_config.get("client_key", None)
        client_secret = vistara_config.get("client_secret", None)

        if not client_id or not client_key or not client_secret:
            return (
                False,
                "vistara client_id or client_key or client_secret "
                "has not been specified in the Salt master config.",
            )
        return True

    return (
        False,
        "vistara config has not been specificed in the Salt master "
        "config. See documentation for this runner.",
    )


def _get_vistara_configuration():
    """
    Return the Vistara configuration read from the master config
    """
    return {
        "client_id": __opts__["vistara"]["client_id"],
        "client_key": __opts__["vistara"]["client_key"],
        "client_secret": __opts__["vistara"]["client_secret"],
    }


def delete_device(name, safety_on=True):
    """
    Deletes a device from Vistara based on DNS name or partial name. By default,
    delete_device will only perform the delete if a single host is returned. Set
    safety_on=False to delete all matches (up to default API search page size)

    CLI Example:

    .. code-block:: bash

        salt-run vistara.delete_device 'hostname-101.mycompany.com'
        salt-run vistara.delete_device 'hostname-101'
        salt-run vistara.delete_device 'hostname-1' safety_on=False

    """

    config = _get_vistara_configuration()
    if not config:
        return False

    access_token = _get_oath2_access_token(
        config["client_key"], config["client_secret"]
    )

    if not access_token:
        return "Vistara access token not available"

    query_string = f"dnsName:{name}"

    devices = _search_devices(query_string, config["client_id"], access_token)

    if not devices:
        return "No devices found"

    device_count = len(devices)

    if safety_on and device_count != 1:
        return (
            "Expected to delete 1 device and found {}. "
            "Set safety_on=False to override.".format(device_count)
        )

    delete_responses = []
    for device in devices:
        device_id = device["id"]
        log.debug(device_id)
        delete_response = _delete_resource(device_id, config["client_id"], access_token)
        if not delete_response:
            return False
        delete_responses.append(delete_response)

    return delete_responses


def _search_devices(query_string, client_id, access_token):

    authstring = f"Bearer {access_token}"

    headers = {
        "Authorization": authstring,
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    params = {"queryString": query_string}

    method = "GET"
    url = f"https://api.vistara.io/api/v2/tenants/{client_id}/devices/search"

    resp = salt.utils.http.query(
        url=url, method=method, header_dict=headers, params=params, opts=__opts__
    )

    respbody = resp.get("body", None)
    if not respbody:
        return False

    respbodydict = salt.utils.json.loads(resp["body"])
    deviceresults = respbodydict["results"]

    return deviceresults


def _delete_resource(device_id, client_id, access_token):

    authstring = f"Bearer {access_token}"

    headers = {
        "Authorization": authstring,
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    method = "DELETE"
    url = "https://api.vistara.io/api/v2/tenants/{}/rtype/DEVICE/resource/{}".format(
        client_id, device_id
    )

    resp = salt.utils.http.query(
        url=url, method=method, header_dict=headers, opts=__opts__
    )

    respbody = resp.get("body", None)
    if not respbody:
        return False

    respbodydict = salt.utils.json.loads(resp["body"])

    return respbodydict


def _get_oath2_access_token(client_key, client_secret):
    """
    Query the vistara API and get an access_token

    """
    if not client_key and not client_secret:
        log.error(
            "client_key and client_secret have not been specified "
            "and are required parameters."
        )
        return False

    method = "POST"
    url = "https://api.vistara.io/auth/oauth/token"
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Accept": "application/json",
    }

    params = {
        "grant_type": "client_credentials",
        "client_id": client_key,
        "client_secret": client_secret,
    }

    resp = salt.utils.http.query(
        url=url, method=method, header_dict=headers, params=params, opts=__opts__
    )

    respbody = resp.get("body", None)

    if not respbody:
        return False

    access_token = salt.utils.json.loads(respbody)["access_token"]
    return access_token



import pytest

import salt.modules.opkg as opkg
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {opkg: {}}


def test_when_os_is_NILinuxRT_and_creation_of_RESTART_CHECK_STATE_PATH_fails_virtual_should_be_False():
    expected_result = (
        False,
        "Error creating /var/lib/salt/restartcheck_state (-whatever): 42",
    )
    with patch.dict(opkg.__grains__, {"os_family": "NILinuxRT"}), patch(
        "os.makedirs", autospec=True, side_effect=OSError("whatever", 42, "boop")
    ):
        result = opkg.__virtual__()
        assert result == expected_result


def test_when_os_is_NILinuxRT_and_creation_is_OK_and_no_files_exist_then_files_should_be_updated():
    patch_grains = patch.dict(opkg.__grains__, {"os_family": "NILinuxRT"})
    patch_makedirs = patch("os.makedirs", autospec=True, return_value=None)
    patch_update_state = patch(
        "salt.modules.opkg._update_nilrt_restart_state", autospec=True
    )
    patch_listdir = patch("os.listdir", return_value=[], autospec=True)
    with patch_grains, patch_makedirs, patch_listdir, patch_update_state as fake_update:
        opkg.__virtual__()

        fake_update.assert_called_once()


def test_when_os_is_NILinuxRT_and_creation_is_OK_and_files_already_exist_then_files_should_not_be_updated():
    patch_grains = patch.dict(opkg.__grains__, {"os_family": "NILinuxRT"})
    patch_makedirs = patch("os.makedirs", autospec=True, return_value=None)
    patch_update_state = patch(
        "salt.modules.opkg._update_nilrt_restart_state", autospec=True
    )
    patch_listdir = patch(
        "os.listdir", return_value=["these", "are", "pretend", "files"], autospec=True
    )
    with patch_grains, patch_makedirs, patch_listdir, patch_update_state as fake_update:
        opkg.__virtual__()

        fake_update.assert_not_called()



import pytest
import tornado.wsgi

import salt.netapi.rest_cherrypy.app
import tests.support.netapi as netapi

cherrypy = pytest.importorskip("cherrypy")


@pytest.fixture
def client_config(client_config, netapi_port, request):
    client_config["rest_cherrypy"] = {"port": netapi_port, "debug": True}
    marker = request.node.get_closest_marker("netapi_client_data")
    if marker is None:
        client_config["netapi_enable_clients"] = []
    else:
        client_config["netapi_enable_clients"] = marker.args[0]
    return client_config


@pytest.fixture
def app(client_config, load_auth, salt_minion):
    app, _, cherry_opts = salt.netapi.rest_cherrypy.app.get_app(client_config)

    return tornado.wsgi.WSGIContainer(
        cherrypy.Application(app, "/", config=cherry_opts)
    )


@pytest.fixture
def client_headers(auth_token, content_type_map):
    return {
        "Content-Type": content_type_map["form"],
    }


# The order of these fixutres matters, app must come before io_loop.
@pytest.fixture
def http_server(app, netapi_port, client_headers, io_loop):
    with netapi.TestsTornadoHttpServer(
        io_loop=io_loop, app=app, port=netapi_port, client_headers=client_headers
    ) as server:
        yield server


@pytest.fixture
def http_client(http_server):
    return http_server.client



"""
Tests for the spm info utility
"""

import shutil

import pytest

from tests.support.case import SPMCase


@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
class SPMInfoTest(SPMCase):
    """
    Validate the spm info command
    """

    def setUp(self):
        self.config = self._spm_config()
        self._spm_build_files(self.config)

    @pytest.mark.slow_test
    def test_spm_info(self):
        """
        test spm build
        """
        self._spm_create_update_repo(self.config)
        install = self.run_spm("install", self.config, "apache")
        get_info = self.run_spm("info", self.config, "apache")

        check_info = ["Supported OSes", "Supported OS", "installing Apache"]
        for info in check_info:
            self.assertIn(info, "".join(get_info))

    def tearDown(self):
        shutil.rmtree(self._tmp_spm)



import shutil

import pytest


@pytest.fixture(scope="package")
def pillar_state_tree(tmp_path_factory):
    _pillar_state_tree = tmp_path_factory.mktemp("pillar")
    try:
        yield _pillar_state_tree
    finally:
        shutil.rmtree(str(_pillar_state_tree), ignore_errors=True)


@pytest.fixture(scope="package")
def extension_modules(tmp_path_factory):
    _extension_modules = tmp_path_factory.mktemp("pillar")
    try:
        yield _extension_modules
    finally:
        shutil.rmtree(str(_extension_modules), ignore_errors=True)


@pytest.fixture(scope="package")
def salt_master(salt_factories, pillar_state_tree, extension_modules):
    config_defaults = {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
        "extension_modules": str(extension_modules),
        "ext_pillar_first": False,
        "ext_pillar": [],
        "decrypt_pillar_default": "gpg",
        "decrypt_pillar_delimiter": ":",
        "decrypt_pillar_renderers": ["gpg"],
    }
    factory = salt_factories.salt_master_daemon(
        "pillar-functional-master", defaults=config_defaults
    )
    return factory


@pytest.fixture(scope="package")
def salt_minion_1(salt_master):
    factory = salt_master.salt_minion_daemon(
        "pillar-functional-minion-1", defaults={"open_mode": True}
    )
    return factory


@pytest.fixture(scope="package")
def salt_minion_2(salt_master):
    factory = salt_master.salt_minion_daemon(
        "pillar-functional-minion-2", defaults={"open_mode": True}
    )
    return factory



import pytest

import salt.crypt as crypt


@pytest.fixture
def key_data():
    return [
        "-----BEGIN PUBLIC KEY-----",
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoe5QSDYRWKyknbVyRrIj",
        "rm1ht5HgKzAVUber0x54+b/UgxTd1cqI6I+eDlx53LqZSH3G8Rd5cUh8LHoGedSa",
        "E62vEiLAjgXa+RdgcGiQpYS8+Z2RvQJ8oIcZgO+2AzgBRHboNWHTYRRmJXCd3dKs",
        "9tcwK6wxChR06HzGqaOTixAuQlegWbOTU+X4dXIbW7AnuQBt9MCib7SxHlscrqcS",
        "cBrRvq51YP6cxPm/rZJdBqZhVrlghBvIpa45NApP5PherGi4AbEGYte4l+gC+fOA",
        "osEBis1V27djPpIyQS4qk3XAPQg6CYQMDltHqA4Fdo0Nt7SMScxJhfH0r6zmBFAe",
        "BQIDAQAB",
        "-----END PUBLIC KEY-----",
    ]


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test__clean_key(key_data, linesep):
    tst_key = linesep.join(key_data)
    chk_key = "\n".join(key_data)
    assert crypt.clean_key(tst_key) == crypt.clean_key(chk_key)


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test__clean_key_mismatch(key_data, linesep):
    tst_key = linesep.join(key_data)
    tst_key = tst_key.replace("5", "4")
    chk_key = "\n".join(key_data)
    assert crypt.clean_key(tst_key) != crypt.clean_key(chk_key)



"""
virst compatibility module for managing VMs on SmartOS
"""

import logging

import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "virt"


def __virtual__():
    """
    Provides virt on SmartOS
    """
    if salt.utils.platform.is_smartos_globalzone() and salt.utils.path.which("vmadm"):
        return __virtualname__
    return (
        False,
        f"{__virtualname__} module can only be loaded on SmartOS compute nodes",
    )


def init(**kwargs):
    """
    Initialize a new VM

    CLI Example:

    .. code-block:: bash

        salt '*' virt.init image_uuid='...' alias='...' [...]
    """
    return __salt__["vmadm.create"](**kwargs)


def list_domains():
    """
    Return a list of virtual machine names on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' virt.list_domains
    """
    data = __salt__["vmadm.list"](keyed=True)
    vms = [
        "UUID                                  TYPE  RAM      STATE             ALIAS"
    ]
    for vm in data:
        vms.append(
            "{vmuuid}{vmtype}{vmram}{vmstate}{vmalias}".format(
                vmuuid=vm.ljust(38),
                vmtype=data[vm]["type"].ljust(6),
                vmram=data[vm]["ram"].ljust(9),
                vmstate=data[vm]["state"].ljust(18),
                vmalias=data[vm]["alias"],
            )
        )
    return vms


def list_active_vms():
    """
    Return a list of uuids for active virtual machine on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' virt.list_active_vms
    """
    return __salt__["vmadm.list"](search="state='running'", order="uuid")


def list_inactive_vms():
    """
    Return a list of uuids for inactive virtual machine on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' virt.list_inactive_vms
    """
    return __salt__["vmadm.list"](search="state='stopped'", order="uuid")


def vm_info(domain):
    """
    Return a dict with information about the specified VM on this CN

    CLI Example:

    .. code-block:: bash

        salt '*' virt.vm_info <domain>
    """
    return __salt__["vmadm.get"](domain)


def start(domain):
    """
    Start a defined domain

    CLI Example:

    .. code-block:: bash

        salt '*' virt.start <domain>
    """
    if domain in list_active_vms():
        raise CommandExecutionError("The specified vm is already running")

    __salt__["vmadm.start"](domain)

    return domain in list_active_vms()


def shutdown(domain):
    """
    Send a soft shutdown signal to the named vm

    CLI Example:

    .. code-block:: bash

        salt '*' virt.shutdown <domain>
    """
    if domain in list_inactive_vms():
        raise CommandExecutionError("The specified vm is already stopped")

    __salt__["vmadm.stop"](domain)

    return domain in list_inactive_vms()


def reboot(domain):
    """
    Reboot a domain via ACPI request

    CLI Example:

    .. code-block:: bash

        salt '*' virt.reboot <domain>
    """
    if domain in list_inactive_vms():
        raise CommandExecutionError("The specified vm is stopped")

    __salt__["vmadm.reboot"](domain)

    return domain in list_active_vms()


def stop(domain):
    """
    Hard power down the virtual machine, this is equivalent to powering off the hardware.

    CLI Example:

    .. code-block:: bash

        salt '*' virt.destroy <domain>
    """
    if domain in list_inactive_vms():
        raise CommandExecutionError("The specified vm is stopped")

    return __salt__["vmadm.delete"](domain)


def vm_virt_type(domain):
    """
    Return VM virtualization type : OS or KVM

    CLI Example:

    .. code-block:: bash

        salt '*' virt.vm_virt_type <domain>
    """
    ret = __salt__["vmadm.lookup"](search=f"uuid={domain}", order="type")
    if not ret:
        raise CommandExecutionError("We can't determine the type of this VM")

    return ret[0]["type"]


def setmem(domain, memory):
    """
    Change the amount of memory allocated to VM.
    <memory> is to be specified in MB.

    Note for KVM : this would require a restart of the VM.

    CLI Example:

    .. code-block:: bash

        salt '*' virt.setmem <domain> 512
    """
    vmtype = vm_virt_type(domain)
    if vmtype == "OS":
        return __salt__["vmadm.update"](vm=domain, max_physical_memory=memory)
    elif vmtype == "LX":
        return __salt__["vmadm.update"](vm=domain, max_physical_memory=memory)
    elif vmtype == "KVM":
        log.warning("Changes will be applied after the VM restart.")
        return __salt__["vmadm.update"](vm=domain, ram=memory)
    else:
        raise CommandExecutionError("Unknown VM type")

    return False


def get_macs(domain):
    """
    Return a list off MAC addresses from the named VM

    CLI Example:

    .. code-block:: bash

        salt '*' virt.get_macs <domain>
    """
    macs = []
    ret = __salt__["vmadm.lookup"](search=f"uuid={domain}", order="nics")
    if not ret:
        raise CommandExecutionError("We can't find the MAC address of this VM")
    else:
        for nic in ret[0]["nics"]:
            macs.append(nic["mac"])
        return macs



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import os
import textwrap

import pytest

import salt.modules.mac_brew_pkg as mac_brew
import salt.utils.pkg
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def TAPS_STRING():
    return "homebrew/dupes\nhomebrew/science\nhomebrew/x11"


@pytest.fixture
def TAPS_LIST():
    return ["homebrew/dupes", "homebrew/science", "homebrew/x11"]


@pytest.fixture
def HOMEBREW_PREFIX():
    return "/opt/homebrew"


@pytest.fixture
def HOMEBREW_BIN(HOMEBREW_PREFIX):
    return HOMEBREW_PREFIX + "/bin/brew"


@pytest.fixture
def configure_loader_modules():
    return {mac_brew: {"__opts__": {"user": MagicMock(return_value="bar")}}}


def custom_call_brew(*cmd, failhard=True):
    result = dict()
    if cmd == ("info", "--json=v2", "--installed"):
        result = {
            "stdout": textwrap.dedent(
                """\
                {
                  "casks": [
                    {
                      "appcast": null,
                      "artifacts": [
                        [
                          "Day-3.0/Day-O.app"
                        ],
                        {
                          "signal": {},
                          "trash": "~/Library/Preferences/com.shauninman.Day-O.plist"
                        }
                      ],
                      "auto_updates": null,
                      "caveats": null,
                      "conflicts_with": null,
                      "container": null,
                      "depends_on": {},
                      "desc": null,
                      "homepage": "https://shauninman.com/archive/2020/04/08/day_o_mac_menu_bar_clock_for_catalina",
                      "installed": "3.0.1",
                      "name": [
                        "Day-O"
                      ],
                      "outdated": false,
                      "sha256": "4963f503c1e47bfa0f8bdbbbe5694d6a7242d298fb44ff68af80d42f1eaebaf9",
                      "token": "day-o",
                      "full_token": "day-o",
                      "tap": "homebrew/cask",
                      "url": "https://shauninman.com/assets/downloads/Day-3.0.zip",
                      "version": "3.0.1"
                    },
                    {
                      "appcast": null,
                      "artifacts": [
                        [
                          "iTerm.app"
                        ],
                        {
                          "signal": {},
                          "trash": [
                            "~/Library/Application Support/iTerm",
                            "~/Library/Application Support/iTerm2",
                            "~/Library/Application Support/com.apple.sharedfilelist/com.apple.LSSharedFileList.ApplicationRecentDocuments/com.googlecode.iterm2.sfl*",
                            "~/Library/Caches/com.googlecode.iterm2",
                            "~/Library/Preferences/com.googlecode.iterm2.plist",
                            "~/Library/Saved Application State/com.googlecode.iterm2.savedState"
                          ]
                        }
                      ],
                      "auto_updates": true,
                      "caveats": null,
                      "conflicts_with": {
                        "cask": [
                          "iterm2-beta"
                        ]
                      },
                      "container": null,
                      "depends_on": {
                        "macos": {
                          ">=": [
                            "10.12"
                          ]
                        }
                      },
                      "desc": "Terminal emulator as alternative to Apple's Terminal app",
                      "homepage": "https://www.iterm2.com/",
                      "installed": "3.4.3",
                      "name": [
                        "iTerm2"
                      ],
                      "outdated": false,
                      "sha256": "9ed73844838bddf797eadf37e5f7da3771308c3f74d38cd422c18eebaaa8f6b9",
                      "token": "iterm2",
                      "full_token": "custom/tap/iterm2",
                      "tap": "custom/tap",
                      "url": "https://iterm2.com/downloads/stable/iTerm2-3_4_3.zip",
                      "version": "3.4.3"
                    },
                    {
                      "token": "discord",
                      "full_token": "discord",
                      "tap": null,
                      "name": [
                        "Discord"
                      ],
                      "desc": "Voice and text chat software",
                      "homepage": "https://discord.com/",
                      "url": "https://dl.discordapp.net/apps/osx/0.0.268/Discord.dmg",
                      "appcast": null,
                      "version": "0.0.268",
                      "versions": {
                      },
                      "installed": "0.0.266",
                      "outdated": false,
                      "sha256": "dfe12315b717ed06ac24d3eaacb700618e96cbb449ed63d2afadcdb70ad09c55",
                      "artifacts": [
                        {
                          "app": [
                            "Discord.app"
                          ]
                        },
                        {
                          "zap": [
                            {
                              "trash": [
                                "~/Library/Application Support/discord",
                                "~/Library/Caches/com.hnc.Discord",
                                "~/Library/Caches/com.hnc.Discord.ShipIt",
                                "~/Library/Cookies/com.hnc.Discord.binarycookies",
                                "~/Library/Preferences/com.hnc.Discord.helper.plist",
                                "~/Library/Preferences/com.hnc.Discord.plist",
                                "~/Library/Saved Application State/com.hnc.Discord.savedState"
                              ]
                            }
                          ]
                        }
                      ],
                      "caveats": null,
                      "depends_on": {
                      },
                      "conflicts_with": null,
                      "container": null,
                      "auto_updates": true
                    }
                  ],
                  "formulae": [
                    {
                      "aliases": [],
                      "bottle": {
                        "stable": {
                          "cellar": ":any",
                          "files": {
                            "arm64_big_sur": {
                              "sha256": "674b3ae41c399f1e8e44c271b0e6909babff9fcd2e04a2127d25e2407ea4dd33",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.arm64_big_sur.bottle.1.tar.gz"
                            },
                            "big_sur": {
                              "sha256": "bf0f8577632af7b878b6425476f5b1ab9c3bf66d65affb0c455048a173a0b6bf",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.big_sur.bottle.1.tar.gz"
                            },
                            "catalina": {
                              "sha256": "820a3c85fcbb63088b160c7edf125d7e55fc2c5c1d51569304499c9cc4b89ce8",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.catalina.bottle.1.tar.gz"
                            },
                            "high_sierra": {
                              "sha256": "dffcffa4ea13e8f0f2b45c5121e529077e135ae9a47254c32182231662ee9b72",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.high_sierra.bottle.1.tar.gz"
                            },
                            "mojave": {
                              "sha256": "71f0e76c5b22e5088426c971d5e795fe67abee7af6c2c4ae0cf4c0eb98ed21ff",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.mojave.bottle.1.tar.gz"
                            },
                            "sierra": {
                              "sha256": "bb4d19dc026c2d72c53eed78eaa0ab982e9fcad2cd2acc6d13e7a12ff658e877",
                              "url": "https://homebrew.bintray.com/bottles/jq-1.6.sierra.bottle.1.tar.gz"
                            }
                          },
                          "prefix": "/usr/local",
                          "rebuild": 1,
                          "root_url": "https://homebrew.bintray.com/bottles"
                        }
                      },
                      "bottle_disabled": false,
                      "build_dependencies": [],
                      "caveats": null,
                      "conflicts_with": [],
                      "dependencies": [
                        "oniguruma"
                      ],
                      "deprecated": false,
                      "deprecation_date": null,
                      "deprecation_reason": null,
                      "desc": "Lightweight and flexible command-line JSON processor",
                      "disable_date": null,
                      "disable_reason": null,
                      "disabled": false,
                      "full_name": "jq",
                      "homepage": "https://stedolan.github.io/jq/",
                      "installed": [
                        {
                          "built_as_bottle": true,
                          "installed_as_dependency": false,
                          "installed_on_request": true,
                          "poured_from_bottle": true,
                          "runtime_dependencies": [
                            {
                              "full_name": "oniguruma",
                              "version": "6.9.6"
                            }
                          ],
                          "used_options": [],
                          "version": "1.6"
                        }
                      ],
                      "keg_only": false,
                      "license": "MIT",
                      "linked_keg": "1.6",
                      "name": "jq",
                      "oldname": null,
                      "optional_dependencies": [],
                      "options": [],
                      "outdated": false,
                      "pinned": false,
                      "recommended_dependencies": [],
                      "requirements": [],
                      "revision": 0,
                      "urls": {
                        "stable": {
                          "revision": null,
                          "tag": null,
                          "url": "https://github.com/stedolan/jq/releases/download/jq-1.6/jq-1.6.tar.gz"
                        }
                      },
                      "uses_from_macos": [],
                      "version_scheme": 0,
                      "versioned_formulae": [],
                      "versions": {
                        "bottle": true,
                        "head": "HEAD",
                        "stable": "1.6"
                      }
                    },
                    {
                      "aliases": [],
                      "bottle": {
                        "stable": {
                          "cellar": ":any",
                          "files": {
                            "arm64_big_sur": {
                              "sha256": "c84206005787304416ed81094bd3a0cdd2ae8eb62649db5a3a44fa14b276d09f",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.arm64_big_sur.bottle.tar.gz"
                            },
                            "big_sur": {
                              "sha256": "4fbd4a9e3eb49c27e83bd125b0e76d386c0e12ae1139d4dc9e31841fb8880a35",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.big_sur.bottle.tar.gz"
                            },
                            "catalina": {
                              "sha256": "2dcc8e0121c934d1e34ffdb37fcd70f0f7b5c2f4755f2f7cbcf360e9e54cb43b",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.catalina.bottle.tar.gz"
                            },
                            "high_sierra": {
                              "sha256": "1491b2b20c40c3cb0b990f520768d7e876e4ab4a7dc1da9994d0150da34ba5c6",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.high_sierra.bottle.tar.gz"
                            },
                            "mojave": {
                              "sha256": "44483961b5d2b535b0ece1936c9d40b4bc7d9c7281646cca0fb476291ab9d4dc",
                              "url": "https://homebrew.bintray.com/bottles/xz-5.2.5.mojave.bottle.tar.gz"
                            }
                          },
                          "prefix": "/usr/local",
                          "rebuild": 0,
                          "root_url": "https://homebrew.bintray.com/bottles"
                        }
                      },
                      "bottle_disabled": false,
                      "build_dependencies": [],
                      "caveats": null,
                      "conflicts_with": [],
                      "dependencies": [],
                      "deprecated": false,
                      "deprecation_date": null,
                      "deprecation_reason": null,
                      "desc": "General-purpose data compression with high compression ratio",
                      "disable_date": null,
                      "disable_reason": null,
                      "disabled": false,
                      "full_name": "xz",
                      "homepage": "https://tukaani.org/xz/",
                      "installed": [
                        {
                          "built_as_bottle": true,
                          "installed_as_dependency": true,
                          "installed_on_request": false,
                          "poured_from_bottle": true,
                          "runtime_dependencies": [],
                          "used_options": [],
                          "version": "5.2.5"
                        }
                      ],
                      "keg_only": false,
                      "license": "GPL-2.0",
                      "linked_keg": "5.2.5",
                      "name": "xz",
                      "oldname": null,
                      "optional_dependencies": [],
                      "options": [],
                      "outdated": false,
                      "pinned": false,
                      "recommended_dependencies": [],
                      "requirements": [],
                      "revision": 0,
                      "urls": {
                        "stable": {
                          "revision": null,
                          "tag": null,
                          "url": "https://downloads.sourceforge.net/project/lzmautils/xz-5.2.5.tar.gz"
                        }
                      },
                      "uses_from_macos": [],
                      "version_scheme": 0,
                      "versioned_formulae": [],
                      "versions": {
                        "bottle": true,
                        "head": null,
                        "stable": "5.2.5"
                      }
                    }
                  ]
                }
                """
            ),
            "stderr": "",
            "retcode": 0,
        }

    return result


def custom_add_pkg(ret, name, newest_version):
    ret[name] = newest_version
    return ret


# '_list_taps' function tests: 1


def test_list_taps(TAPS_STRING, TAPS_LIST, HOMEBREW_BIN):
    """
    Tests the return of the list of taps
    """
    mock_taps = MagicMock(return_value={"stdout": TAPS_STRING, "retcode": 0})
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch.dict(
            mac_brew.__salt__,
            {"file.get_user": mock_user, "cmd.run_all": mock_taps, "cmd.run": mock_cmd},
        ):
            assert mac_brew._list_taps() == TAPS_LIST


# '_tap' function tests: 3


def test_tap_installed(TAPS_LIST):
    """
    Tests if tap argument is already installed or not
    """
    with patch(
        "salt.modules.mac_brew_pkg._list_taps", MagicMock(return_value=TAPS_LIST)
    ):
        assert mac_brew._tap("homebrew/science")


def test_tap_failure(HOMEBREW_BIN):
    """
    Tests if the tap installation failed
    """
    mock_failure = MagicMock(return_value={"stdout": "", "stderr": "", "retcode": 1})
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch.dict(
            mac_brew.__salt__,
            {
                "cmd.run_all": mock_failure,
                "file.get_user": mock_user,
                "cmd.run": mock_cmd,
            },
        ), patch("salt.modules.mac_brew_pkg._list_taps", MagicMock(return_value={})):
            assert not mac_brew._tap("homebrew/test")


def test_tap(TAPS_LIST, HOMEBREW_BIN):
    """
    Tests adding unofficial GitHub repos to the list of brew taps
    """
    mock_failure = MagicMock(return_value={"retcode": 0})
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch.dict(
            mac_brew.__salt__,
            {
                "cmd.run_all": mock_failure,
                "file.get_user": mock_user,
                "cmd.run": mock_cmd,
            },
        ), patch(
            "salt.modules.mac_brew_pkg._list_taps", MagicMock(return_value=TAPS_LIST)
        ):
            assert mac_brew._tap("homebrew/test")


# 'homebrew_prefix' function tests: 4


def test_homebrew_prefix_env(HOMEBREW_PREFIX):
    """
    Test the path to the homebrew prefix by looking
    at the HOMEBREW_PREFIX environment variable.
    """
    mock_env = os.environ.copy()
    mock_env["HOMEBREW_PREFIX"] = HOMEBREW_PREFIX

    with patch.dict(os.environ, mock_env):
        assert mac_brew.homebrew_prefix() == HOMEBREW_PREFIX


def test_homebrew_prefix_command(HOMEBREW_PREFIX, HOMEBREW_BIN):
    """
    Test the path to the homebrew prefix by running
    the brew --prefix command when the HOMEBREW_PREFIX
    environment variable is not set.
    """
    mock_env = os.environ.copy()
    if "HOMEBREW_PREFIX" in mock_env:
        del mock_env["HOMEBREW_PREFIX"]

    with patch.dict(os.environ, mock_env):
        with patch(
            "salt.modules.cmdmod.run", MagicMock(return_value=HOMEBREW_PREFIX)
        ), patch("salt.modules.file.get_user", MagicMock(return_value="foo")), patch(
            "salt.modules.mac_brew_pkg._homebrew_os_bin",
            MagicMock(return_value=HOMEBREW_BIN),
        ):
            assert mac_brew.homebrew_prefix() == HOMEBREW_PREFIX


def test_homebrew_prefix_returns_none():
    """
    Tests that homebrew_prefix returns None when
    all attempts fail.
    """

    mock_env = os.environ.copy()
    if "HOMEBREW_PREFIX" in mock_env:
        del mock_env["HOMEBREW_PREFIX"]

    with patch.dict(os.environ, mock_env, clear=True):
        with patch(
            "salt.modules.mac_brew_pkg._homebrew_os_bin", MagicMock(return_value=None)
        ):
            assert mac_brew.homebrew_prefix() is None


def test_homebrew_prefix_returns_none_even_with_execution_errors():
    """
    Tests that homebrew_prefix returns None when
    all attempts fail even with command execution errors.
    """

    mock_env = os.environ.copy()
    if "HOMEBREW_PREFIX" in mock_env:
        del mock_env["HOMEBREW_PREFIX"]

    with patch.dict(os.environ, mock_env, clear=True):
        with patch(
            "salt.modules.cmdmod.run", MagicMock(side_effect=CommandExecutionError)
        ), patch(
            "salt.modules.mac_brew_pkg._homebrew_os_bin",
            MagicMock(return_value=None),
        ):
            assert mac_brew.homebrew_prefix() is None


# '_homebrew_os_bin' function tests: 1


def test_homebrew_os_bin_fallback_apple_silicon():
    """
    Test the path to the homebrew executable for Apple Silicon.

    This test checks that even if the PATH does not contain
    the default Homebrew's prefix for the Apple Silicon
    architecture, it is appended.
    """

    # Ensure Homebrew's prefix for Apple Silicon is not present in the PATH
    mock_env = os.environ.copy()
    mock_env["PATH"] = "/usr/local/bin:/usr/bin"

    apple_silicon_homebrew_path = "/opt/homebrew/bin"
    apple_silicon_homebrew_bin = f"{apple_silicon_homebrew_path}/brew"

    def mock_utils_path_which(*args):
        if apple_silicon_homebrew_path in os.environ.get("PATH", "").split(
            os.path.pathsep
        ):
            return apple_silicon_homebrew_bin
        return None

    with patch("salt.utils.path.which", mock_utils_path_which):
        assert mac_brew._homebrew_os_bin() == apple_silicon_homebrew_bin


# '_homebrew_bin' function tests: 1


def test_homebrew_bin(HOMEBREW_PREFIX, HOMEBREW_BIN):
    """
    Tests the path to the homebrew binary
    """
    mock_path = MagicMock(return_value=HOMEBREW_PREFIX)
    with patch("salt.modules.mac_brew_pkg.homebrew_prefix", mock_path):
        assert mac_brew._homebrew_bin() == HOMEBREW_BIN


# 'list_pkgs' function tests: 2
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_list_pkgs_removed():
    """
    Tests removed implementation
    """
    assert mac_brew.list_pkgs(removed=True) == {}


def test_list_pkgs_versions_true():
    """
    Tests if pkg.list_pkgs is already in context and is a list
    """
    mock_context = {"foo": ["bar"]}
    with patch.dict(mac_brew.__context__, {"pkg.list_pkgs": mock_context}):
        assert mac_brew.list_pkgs(versions_as_list=True) == mock_context


def test_list_pkgs_homebrew_cask_pakages():
    """
    Tests if pkg.list_pkgs list properly homebrew cask packages
    """
    expected_pkgs = {
        "homebrew/cask/day-o": "3.0.1",
        "day-o": "3.0.1",
        "homebrew/cask/discord": "0.0.266",
        "discord": "0.0.266",
        "custom/tap/iterm2": "3.4.3",
        "iterm2": "3.4.3",
        "jq": "1.6",
        "xz": "5.2.5",
    }

    with patch("salt.modules.mac_brew_pkg._call_brew", custom_call_brew), patch.dict(
        mac_brew.__salt__,
        {
            "pkg_resource.add_pkg": custom_add_pkg,
            "pkg_resource.sort_pkglist": MagicMock(),
        },
    ):
        assert mac_brew.list_pkgs(versions_as_list=True) == expected_pkgs


def test_list_pkgs_no_context():
    """
    Tests removed implementation
    """

    expected_pkgs = {
        "zsh": "5.7.1",
        "homebrew/cask/macvim": "8.1.151",
        "homebrew/cask-fonts/font-firacode-nerd-font": "2.0.0",
    }

    with patch("salt.modules.mac_brew_pkg._call_brew", custom_call_brew), patch.dict(
        mac_brew.__salt__,
        {
            "pkg_resource.add_pkg": custom_add_pkg,
            "pkg_resource.sort_pkglist": MagicMock(),
        },
    ), patch.object(mac_brew, "_list_pkgs_from_context") as list_pkgs_context_mock:
        pkgs = mac_brew.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()

        pkgs = mac_brew.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()


# 'version' function tests: 1


def test_version():
    """
    Tests version name returned
    """
    mock_version = MagicMock(return_value="0.1.5")
    with patch.dict(mac_brew.__salt__, {"pkg_resource.version": mock_version}):
        assert mac_brew.version("foo") == "0.1.5"


# 'latest_version' function tests: 0
# It has not been fully implemented

# 'remove' function tests: 1
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_remove():
    """
    Tests if package to be removed exists
    """
    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg.list_pkgs", return_value={"test": "0.1.5"}
    ), patch.dict(mac_brew.__salt__, {"pkg_resource.parse_targets": mock_params}):
        assert mac_brew.remove("foo") == {}


# 'refresh_db' function tests: 2


def test_refresh_db_failure(HOMEBREW_BIN):
    """
    Tests an update of homebrew package repository failure
    """
    mock_user = MagicMock(return_value="foo")
    mock_failure = MagicMock(return_value={"stdout": "", "stderr": "", "retcode": 1})
    with patch("salt.utils.path.which", MagicMock(return_value="/usr/local/bin/brew")):
        with patch.dict(
            mac_brew.__salt__, {"file.get_user": mock_user, "cmd.run_all": mock_failure}
        ), patch(
            "salt.modules.mac_brew_pkg._homebrew_bin",
            MagicMock(return_value=HOMEBREW_BIN),
        ):
            with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
                pytest.raises(CommandExecutionError, mac_brew.refresh_db)


def test_refresh_db(HOMEBREW_BIN):
    """
    Tests a successful update of homebrew package repository
    """
    mock_user = MagicMock(return_value="foo")
    mock_success = MagicMock(return_value={"retcode": 0})
    with patch("salt.utils.path.which", MagicMock(return_value=HOMEBREW_BIN)):
        with patch.dict(
            mac_brew.__salt__, {"file.get_user": mock_user, "cmd.run_all": mock_success}
        ), patch(
            "salt.modules.mac_brew_pkg._homebrew_bin",
            MagicMock(return_value=HOMEBREW_BIN),
        ):
            with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
                assert mac_brew.refresh_db()


# 'install' function tests: 1
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_install():
    """
    Tests if package to be installed exists
    """
    mock_params = MagicMock(return_value=[None, None])
    with patch.dict(mac_brew.__salt__, {"pkg_resource.parse_targets": mock_params}):
        assert mac_brew.install("name=foo") == {}


# "hold" function tests: 2
# Only tested a few basics
# Full functionality should be tested in integration phase


def test_hold(HOMEBREW_BIN):
    """
    Tests holding if package is installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {"new": "hold", "old": "install"},
            "comment": "Package foo is now being held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch(
            "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "pkg_resource.parse_targets": mock_params,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert mac_brew.hold("foo") == _expected


def test_hold_not_installed(HOMEBREW_BIN):
    """
    Tests holding if package is not installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo does not have a state.",
            "name": "foo",
            "result": False,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "pkg_resource.parse_targets": mock_params,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert mac_brew.hold("foo") == _expected


def test_hold_pinned():
    """
    Tests holding if package is already pinned
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo is already set to be held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
    ), patch(
        "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
    ), patch.dict(
        mac_brew.__salt__,
        {
            "file.get_user": mock_user,
            "pkg_resource.parse_targets": mock_params,
            "cmd.run_all": mock_cmd_all,
            "cmd.run": mock_cmd,
        },
    ):
        assert mac_brew.hold("foo") == _expected


# "unhold" function tests: 2
# Only tested a few basics
# Full functionality should be tested in integration phase
def test_unhold(HOMEBREW_BIN):
    """
    Tests unholding if package is installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {"new": "install", "old": "hold"},
            "comment": "Package foo is no longer being held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch(
            "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
        ), patch(
            "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "pkg_resource.parse_targets": mock_params,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert mac_brew.unhold("foo") == _expected


def test_unhold_not_installed():
    """
    Tests unholding if package is not installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo does not have a state.",
            "name": "foo",
            "result": False,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch(
        "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
    ), patch.dict(
        mac_brew.__salt__,
        {
            "file.get_user": mock_user,
            "pkg_resource.parse_targets": mock_params,
            "cmd.run_all": mock_cmd_all,
            "cmd.run": mock_cmd,
        },
    ):
        assert mac_brew.unhold("foo") == _expected


def test_unhold_not_pinned():
    """
    Tests unholding if package is not installed
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}
    )
    _expected = {
        "foo": {
            "changes": {},
            "comment": "Package foo is already set not to be held.",
            "name": "foo",
            "result": True,
        }
    }

    mock_params = MagicMock(return_value=({"foo": None}, "repository"))
    with patch(
        "salt.modules.mac_brew_pkg.list_pkgs", return_value={"foo": "0.1.5"}
    ), patch("salt.modules.mac_brew_pkg._list_pinned", return_value=[]), patch.dict(
        mac_brew.__salt__,
        {
            "file.get_user": mock_user,
            "pkg_resource.parse_targets": mock_params,
            "cmd.run_all": mock_cmd_all,
            "cmd.run": mock_cmd,
        },
    ):
        assert mac_brew.unhold("foo") == _expected


def test_info_installed(HOMEBREW_BIN):
    """
    Tests info_installed method
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={
            "pid": 12345,
            "retcode": 0,
            "stderr": "",
            "stdout": textwrap.dedent(
                """\
                {
                  "formulae": [
                    {
                      "name": "salt",
                      "full_name": "cdalvaro/tap/salt",
                      "tap": "cdalvaro/tap",
                      "aliases": []
                    },
                    {
                      "name": "vim",
                      "full_name": "vim",
                      "tap": "homebrew/core",
                      "aliases": []
                    }
                  ],
                  "casks": [
                    {
                      "token": "visual-studio-code",
                      "full_token": "visual-studio-code",
                      "tap": null,
                      "name": [
                        "MicrosoftVisualStudioCode",
                        "VSCode"
                      ]
                    }
                  ]
                }
             """
            ),
        }
    )
    _expected = {
        "cdalvaro/tap/salt": {
            "name": "salt",
            "full_name": "cdalvaro/tap/salt",
            "tap": "cdalvaro/tap",
            "aliases": [],
        },
        "vim": {
            "name": "vim",
            "full_name": "vim",
            "tap": "homebrew/core",
            "aliases": [],
        },
        "visual-studio-code": {
            "token": "visual-studio-code",
            "full_token": "visual-studio-code",
            "tap": None,
            "name": ["MicrosoftVisualStudioCode", "VSCode"],
        },
    }

    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch(
            "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert (
                mac_brew.info_installed(
                    "cdalvaro/tap/salt", "vim", "visual-studio-code"
                )
                == _expected
            )


def test_list_upgrades(HOMEBREW_BIN):
    """
    Tests list_upgrades method
    """
    mock_user = MagicMock(return_value="foo")
    mock_cmd = MagicMock(return_value="")
    mock_cmd_all = MagicMock(
        return_value={
            "pid": 12345,
            "retcode": 0,
            "stderr": "",
            "stdout": textwrap.dedent(
                """\
                {
                  "formulae": [
                    {
                      "name": "cmake",
                      "installed_versions": ["3.19.3"],
                      "current_version": "3.19.4",
                      "pinned": false,
                      "pinned_version": null
                    },
                    {
                      "name": "fzf",
                      "installed_versions": ["0.25.0"],
                      "current_version": "0.25.1",
                      "pinned": false,
                      "pinned_version": null
                    }
                  ],
                  "casks": [
                    {
                      "name": "ksdiff",
                      "installed_versions": "2.2.0,122",
                      "current_version": "2.3.6,123-jan-18-2021"
                    }
                  ]
                }
                """
            ),
        }
    )
    _expected = {
        "cmake": "3.19.4",
        "fzf": "0.25.1",
        "ksdiff": "2.3.6,123-jan-18-2021",
    }

    with patch(
        "salt.modules.mac_brew_pkg._homebrew_bin", MagicMock(return_value=HOMEBREW_BIN)
    ):
        with patch("salt.modules.mac_brew_pkg.list_pkgs", return_value={}), patch(
            "salt.modules.mac_brew_pkg._list_pinned", return_value=["foo"]
        ), patch.dict(
            mac_brew.__salt__,
            {
                "file.get_user": mock_user,
                "cmd.run_all": mock_cmd_all,
                "cmd.run": mock_cmd,
            },
        ):
            assert (
                mac_brew.list_upgrades(refresh=False, include_casks=True) == _expected
            )



import pytest

import salt.states.gpg as gpg
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {gpg: {"__opts__": {"test": False}}}


@pytest.fixture
def keys_list():
    return [
        {
            "keyid": "A",
            "fingerprint": "A",
            "uids": ["Key A"],
            "created": "2010-04-01",
            "keyLength": "4096",
            "ownerTrust": "Ultimately Trusted",
            "trust": "Ultimately Trusted",
        },
        {
            "keyid": "B",
            "fingerprint": "B",
            "uids": ["Key B"],
            "created": "2017-03-06",
            "keyLength": "4096",
            "ownerTrust": "Unknown",
            "trust": "Fully Trusted",
        },
        {
            "keyid": "C",
            "fingerprint": "C",
            "uids": ["Key C"],
            "expires": "2022-06-24",
            "created": "2018-06-24",
            "keyLength": "4096",
            "ownerTrust": "Unknown",
            "trust": "Expired",
        },
        {
            "keyid": "D",
            "fingerprint": "D",
            "uids": ["Key D"],
            "created": "2018-01-18",
            "keyLength": "3072",
            "ownerTrust": "Unknown",
            "trust": "Unknown",
        },
        {
            "keyid": "E",
            "fingerprint": "E",
            "uids": ["Key E"],
            "expires": "2222-11-18",
            "created": "2019-11-20",
            "keyLength": "4096",
            "ownerTrust": "Unknown",
            "trust": "Unknown",
        },
    ]


@pytest.fixture
def gpg_list_keys(request, keys_list):
    list_ = Mock(spec="salt.modules.gpg.list_keys")
    list_.return_value = getattr(request, "param", keys_list)
    with patch.dict(gpg.__salt__, {"gpg.list_keys": list_}):
        yield list_


@pytest.fixture
def gpg_trust(request):
    trust = Mock(spec="salt.modules.gpg.trust_key")
    trust.return_value = getattr(
        request,
        "param",
        {"res": True, "message": "Setting ownership trust to Marginally"},
    )
    with patch.dict(gpg.__salt__, {"gpg.trust_key": trust}):
        yield trust


@pytest.fixture()
def gpg_receive(request):
    recv = Mock(spec="salt.modules.gpg.receive_keys")
    recv.return_value = getattr(
        request, "param", {"res": True, "message": ["Key new added to keychain"]}
    )
    with patch.dict(gpg.__salt__, {"gpg.receive_keys": recv}):
        yield recv


@pytest.mark.usefixtures("gpg_list_keys")
@pytest.mark.parametrize(
    "gpg_trust,expected",
    [
        ({"res": True, "message": "Setting ownership trust to Marginally"}, True),
        ({"res": False, "message": "KeyID A not in GPG keychain"}, False),
    ],
    indirect=["gpg_trust"],
)
def test_gpg_present_trust_change(gpg_receive, gpg_trust, expected):
    ret = gpg.present("A", trust="marginally")
    assert ret["result"] is expected
    assert bool(ret["changes"]) is expected
    gpg_trust.assert_called_once()
    gpg_receive.assert_not_called()


@pytest.mark.usefixtures("gpg_list_keys")
@pytest.mark.parametrize(
    "gpg_receive,expected",
    [
        ({"res": True, "message": ["Key new added to keychain"]}, True),
        (
            {
                "res": False,
                "message": [
                    "Something went wrong during gpg call: gpg: key new: no user ID"
                ],
            },
            False,
        ),
    ],
    indirect=["gpg_receive"],
)
def test_gpg_present_new_key(gpg_receive, gpg_trust, expected):
    ret = gpg.present("new")
    assert ret["result"] is expected
    assert bool(ret["changes"]) is expected
    gpg_receive.assert_called_once()
    gpg_trust.assert_not_called()


@pytest.mark.usefixtures("gpg_list_keys")
@pytest.mark.parametrize(
    "gpg_trust,expected",
    [
        ({"res": True, "message": "Setting ownership trust to Marginally"}, True),
        ({"res": False, "message": "KeyID A not in GPG keychain"}, False),
    ],
    indirect=["gpg_trust"],
)
@pytest.mark.usefixtures("gpg_list_keys")
def test_gpg_present_new_key_and_trust(gpg_receive, gpg_trust, expected):
    ret = gpg.present("new", trust="marginally")
    assert ret["result"] is expected
    # the key is always marked as added
    assert ret["changes"]
    gpg_receive.assert_called_once()
    gpg_trust.assert_called_once()


@pytest.mark.usefixtures("gpg_list_keys")
@pytest.mark.parametrize("key,trust", [("new", None), ("A", "marginally")])
def test_gpg_present_test_mode_no_changes(gpg_receive, gpg_trust, key, trust):
    with patch.dict(gpg.__opts__, {"test": True}):
        ret = gpg.present(key, trust=trust)
        gpg_receive.assert_not_called()
        gpg_trust.assert_not_called()
        assert ret["result"] is None
        assert ret["changes"]
